#ifndef GEN_REC_INTER
#define GEN_REC_INTER

// compiler dependencies
#if (__SUNPRO_CC)
#define typename
#endif

// includes
#include <vector.h>
#include <pair.h>
#include <algo.h>
#include <gen_interval_tree.h>
#include <STL_extensions.h>

// class GenRecInters_user_compare expects compare function objects from user
template < class ForwardIt,
           class LLXDA, class LLYDA,
           class URXDA, class URYDA,
           class CompareX, class CompareY >
class GenRecInters_user_compare
{   
public:
    typedef typename LLXDA::value_type value_type;
private:
    // the sls
    typedef GenIntervalTree_user_compare<ForwardIt, LLYDA, URYDA, CompareY>
    Sweep_line_status;
    typedef typename Sweep_line_status::Interval_position Sls_position;

    // an event point
    struct Rectangle_point;
    friend struct Rectangle_point;
    struct Rectangle_point
    {
	value_type x_coord;
	ForwardIt rectangle;
	Sls_position* sls_pos;
	bool rec_begin;
	
	// constructors
	Rectangle_point() {}
	Rectangle_point(value_type x,
			ForwardIt rec,
			Sls_position* pos_it,
			bool begin = false)
	    : x_coord(x), rectangle(rec), sls_pos(pos_it), rec_begin(begin)
	{}
    };
    
    // comparing two rectangle points to sort the eps
    class Rectangle_point_compare;
    friend class Rectangle_point_compare;
    class Rectangle_point_compare
    {
    public:
	Rectangle_point_compare(CompareX const &cmp) : my_compare(cmp)
	{}
	bool operator()(Rectangle_point const &rp1, Rectangle_point const &rp2)
	{
	    return my_compare(make_pair(rp1.x_coord, rp1.rectangle),
			      make_pair(rp2.x_coord, rp2.rectangle));
	}
    private:
	CompareX my_compare;
    };
    
    // the eps
  typedef vector<Rectangle_point> Event_point_schedule;
  typedef typename Event_point_schedule::iterator Eps_it;
public:
    // type of the iterator used to iterate over the solution
    typedef typename Sweep_line_status::Iterator Rectangle_it;
    
    // class constructor
    GenRecInters_user_compare(ForwardIt beg, ForwardIt end,
				    LLXDA lxda = LLXDA(),
				    LLYDA lyda = LLYDA(),
				    URXDA hxda = URXDA(),
				    URYDA hyda = URYDA(),
				    CompareX comp_x = CompareX(),
				    CompareY comp_y = CompareY())
	: my_lowX(lxda), my_lowY(lyda), my_highX(hxda), my_highY(hyda),
	  my_state(none)
    {
	init_eps(beg, end);
	__inplace_stable_sort(eps.begin(), eps.end(),
			      Rectangle_point_compare(comp_x));  
        my_current_ep = eps.begin();
        sls = new Sweep_line_status(beg, end, lyda, hyda, comp_y);
    }
  
    // class destructor
    virtual ~GenRecInters_user_compare()
    {
        delete[] sls_positions;
        delete sls;
    }

    // states of the algorithm
    enum Algo_state { none, rectangle_start, rectangle_end, done };  
    bool valid() const { return my_state != done;}  
    ForwardIt current() { return (*my_current_ep).rectangle; }

    // iterator used to iterate over the solution
    Rectangle_it begin()
    {
	return sls->begin((*my_current_ep).rectangle);
    }  
    Rectangle_it end()
    {
	return sls->end();
    }

    // loop kernel
    Algo_state step()
    {  
	// move iterator to next event point (if not at beginning or end)
	if ( (my_state != none) && (my_current_ep != eps.end()) )
	{
	    // if the last event point corresponded to a rectangle's RIGHT edge
	    // => delete rectangle from sls
	    if (my_state == rectangle_end)
	    {
		sls->erase(*((*my_current_ep).sls_pos));
	    }
	    my_current_ep++;
	    while ( (my_current_ep != eps.end()) &&
		    !((*my_current_ep).rec_begin) &&
		    sls->is_erased(*((*my_current_ep).sls_pos)) )
	    {
		my_current_ep++;
	    }
        }
  
        // stop if sweep is finished
        if (my_current_ep == eps.end()) my_state = done;
	
        // if the current event point corresponds to a rectangle's LEFT edge
        // => insert rectangle into sls
        else if ((*my_current_ep).rec_begin)
        {
	    *((*my_current_ep).sls_pos) =
		sls->insert((*my_current_ep).rectangle);
	    my_state = rectangle_start;
        }
        // if the current event point corresponded to a rectangle's RIGHT edge
        // => deletion of rectangle from sls is delayed (see above)
        else
        {
	    my_state = rectangle_end;
        }
        return my_state;
    }
    
    void erase_current_conflicts()
    {
        Rectangle_it current = begin(), next;
        while (current != end())
        {
	    next = current;
	    ++next;
	    sls->erase(current);
	    current = next;
        }
    }
    
    // current number of conflicts
    int current_num_of_conflicts()
    {
	int i = 0;
	Rectangle_it my_begin = begin(), my_end = end();
	while (my_begin != my_end)	
	{
	    i++;
	    my_begin++;
	}	
	return i;
    }
    
    // total number of conflicts
    int num_of_conflicts()
    {
	int i = 0;
	while (my_state != done)
	{ 
	    step();
	    if (my_state == rectangle_start)	   
		i += current_num_of_conflicts();	   
	}
	return i;
    }
private:
    Sweep_line_status* sls;
    Event_point_schedule eps;
    Sls_position* sls_positions;
    Eps_it my_current_ep;
    LLXDA my_lowX;
    LLYDA my_lowY;
    URXDA my_highX;
    URYDA my_highY;
    Algo_state my_state;   
    
    void init_eps(ForwardIt const &begin, ForwardIt const &end)
    {
	// number of input rectangles
	int len = lower_distance(begin, end);
	sls_positions = new Sls_position[len];
	Sls_position* pos_it = sls_positions;
	// list of endpoints used to initialize the IntervalTree
	eps.reserve(2*len);
	for (ForwardIt my_current = begin; my_current != end; ++my_current)
	{
	    value_type x1 = get(my_lowX, my_current);
	    value_type x2 = get(my_highX, my_current);
	    
	    eps.push_back(Rectangle_point(x1, my_current, pos_it, true));
	    eps.push_back(Rectangle_point(x2, my_current, pos_it++));
	}
    }
};

// compare function object used by class GenRecInters
template<class ForwardIt, class LLXDA>
class Compare_x
{
  typedef typename LLXDA::value_type value_type;
  typedef pair<value_type, ForwardIt> Pair;
public:
  Compare_x(LLXDA lxda = LLXDA()): my_lowX(lxda)
  {}
  bool operator()(const Pair& p, const Pair& q)
  {
    bool before = true;
    value_type p_x = p.first;
    value_type q_x = q.first;
    if (p_x == q_x)
    {
      value_type p_x1 = get(my_lowX, p.second);
      // p is the x-coordinate of the left edge of a rectangle
      if (p_x == p_x1)
        before = false;
    }
    else before = (p_x < q_x);
    return before;
  }
private:
  LLXDA my_lowX;
};

template <class DA>
class Compare_y
{
  typedef typename DA::value_type value_type;
public:
  bool operator()(value_type const v1, value_type const v2)
  {
    return (v1 < v2);
  }
};

// class GenRecInters
template < class ForwardIt,
           class LLXDA, class LLYDA,
           class URXDA, class URYDA >
class GenRecInters
: public GenRecInters_user_compare< ForwardIt,
                                          LLXDA, LLYDA,
                                          URXDA, URYDA,
                                          Compare_x<ForwardIt, LLXDA>,
                                          Compare_y<LLXDA> >
{
  typedef Compare_x<ForwardIt, LLXDA> CompareX;
  typedef Compare_y<LLXDA> CompareY;
  typedef GenRecInters_user_compare<ForwardIt, LLXDA, LLYDA,
    URXDA, URYDA, CompareX, CompareY> Rectangle_intersection;
public:
  GenRecInters( ForwardIt beg, ForwardIt end,
                              LLXDA lxda = LLXDA(),
                              LLYDA lyda = LLYDA(),
                              URXDA hxda = URXDA(),
                              URYDA hyda = URYDA() )
  : Rectangle_intersection(beg, end, lxda, lyda, hxda, hyda)
  {}
};

#endif
