#ifndef _UGRAPH
#define _UGRAPH

#include <list.h>
#include <STL_extensions.h>

template <class NodeData, class EdgeData>
class Edge_type;

template <class NodeData, class EdgeData>
class Node_type
{  
  typedef Edge_type<NodeData, EdgeData>* Edge_ptr;
  typedef Node_type<NodeData, EdgeData>* Node_ptr;  
  typedef typename list<Edge_ptr>::iterator Edge_list_it;
  typedef list<Edge_list_it> Adj_edges;  
  typedef typename list<Node_ptr>::iterator Node_list_it;
public:
  typedef typename Adj_edges::iterator Adj_edge_it;   
  Node_type(NodeData node_inf): inf(node_inf)
  {}
  Node_type()
  {}
  NodeData& info()
  {
    return inf;
  }
  
  int degree()
  {
    return adj_edges.size();
  }
  
  Adj_edge_it adj_edges_begin()
  {
    return adj_edges.begin();
  }
  
  Adj_edge_it adj_edges_end()
  {
    return adj_edges.end();
  }
  
  Adj_edge_it add_edge(Edge_list_it it_edge)
  {
    return Adj_edge_it(adj_edges.insert(adj_edges.end(), it_edge));
  }
  
  void del_edge(Adj_edge_it it)
  {
    adj_edges.erase(it);
  }
  
  void del_adj_edges(list<Edge_ptr>& edge_list)
  {
    Adj_edge_it end = adj_edges.end();
    for(Adj_edge_it current = adj_edges.begin(); current != end; current++)
    {
      delete **current;
      edge_list.erase(*current);
    }
  }
private: 
  NodeData inf;
  Adj_edges adj_edges;                      
};

template <class NodeData, class EdgeData>
class Edge_type
{  
  typedef Node_type<NodeData, EdgeData>* Node_ptr;
  typedef Edge_type<NodeData, EdgeData>* Edge_ptr;  
  typedef typename list<Edge_ptr>::iterator Edge_list_it;  
  typedef list<Edge_list_it> Adj_edges;
  typedef typename Adj_edges::iterator Adj_edge_it; 
public:
  typedef typename list<Node_ptr>::iterator Node_list_it;
  Edge_type()
  {}
  
  Edge_type(Node_list_it n1, Node_list_it n2, EdgeData edge_inf)
  : inf(edge_inf), v(n1), w(n2)
  {}
  
  ~Edge_type()
  {
    (**v).del_edge(v_edge);
    (**w).del_edge(w_edge);
  }

  EdgeData& info()
  {
    return inf;
  }
  
  Node_list_it& opposite(Node_list_it& n)
  {
    return (n==v) ?  w : v;
  }
  
  NodeData& get_opposite_node_info(Node_list_it& n)
  {
    Node_list_it other = opposite(n);
    return (**other).info();
  }
  
  void set_edge_pos(Edge_list_it it)
  {
    edge_pos = it;
    v_edge = (**v).add_edge(it);
    w_edge = (**w).add_edge(it);
  }
  
  void hide()
  {
    (**v).del_edge(v_edge);
    (**w).del_edge(w_edge);
  }
  
  void restore()
  {
    v_edge = (**v).add_edge(edge_pos);
    w_edge = (**w).add_edge(edge_pos);
  }
private:
  EdgeData inf;
  Edge_list_it edge_pos;
  Node_list_it v, w;
  Adj_edge_it v_edge, w_edge;                             
};

template <class T>
struct DA
{
  typedef T value_type;
};

template <class T, class It>
inline  T get(DA<T> da, It& it)
{
  return (*it).info();
}

template <class T>
struct AdjEdgeDA
{
  typedef T value_type;
};

template <class T, class It>
T get(AdjEdgeDA<T> da, It& it)
{
  return (**it).info();
}

template <class T, class NodeIt>
struct AdjNodeDA
{
  typedef T value_type;

  AdjNodeDA()
  {}

  AdjNodeDA(NodeIt n): node(n)
  {}

  NodeIt get_node()
  {
    return node;
  }
private:
  NodeIt node;
};

template <class T, class EdgeIt, class NodeIt>
T get(AdjNodeDA<T, NodeIt> da, EdgeIt& it)
{
  NodeIt node = da.get_node();
  return (**it).get_opposite_node_info(node);
}

template <class NodeData, class EdgeData>
class Ugraph
{  
  typedef Node_type<NodeData, EdgeData> Node;
  typedef Edge_type<NodeData, EdgeData> Edge;
  typedef Node* Node_ptr;
  typedef Edge* Edge_ptr;
  typedef list<Node_ptr> Node_list;
  typedef list<Edge_ptr> Edge_list;
  typedef typename Node_list::iterator Node_list_it;
  typedef typename Edge_list::iterator Edge_list_it;
  typedef typename list<Node_list_it>::iterator Adj_node_list_it;
  typedef typename list<Edge_list_it>::iterator Adj_edge_list_it;
public:
  typedef pair<NodeData, EdgeData> value_type;
  typedef member_access_it<Node_list_it, DA<NodeData> > Node_it;
  typedef member_access_it<Edge_list_it, DA<EdgeData> > Edge_it;
  typedef member_access_it<Adj_edge_list_it, AdjEdgeDA<EdgeData> >
    Adj_edge_it;
  typedef AdjNodeDA<NodeData, Node_list_it> Adj_node_da;
  typedef member_access_it<Adj_edge_list_it, Adj_node_da> Adj_node_it;

  ~Ugraph()
  {
    Node_list_it current = node_list.begin();
    Node_list_it end = node_list.end();
    for ( ; current != end; current++)
      delete (*current);
  }

  Node_it new_node(NodeData node_inf)
  {
    return Node_it(node_list.insert(node_list.end(), new Node(node_inf)));
  }
  
  void del_node(Node_it n)
  {
    Node_list_it n_it = n.get_iterator();
    (**n_it).del_adj_edges(edge_list);
    delete *n_it;
    node_list.erase(n_it);
  }
  
  Edge_it new_edge(Node_it v, Node_it w, EdgeData edge_inf)
  {
    Edge_list_it it = edge_list.insert(
      edge_list.end(), new Edge(v.get_iterator(), w.get_iterator(), edge_inf));
    (**it).set_edge_pos(it);
    return Edge_it(it);
  }
  
  void del_edge(Edge_it e)
  {
    Edge_list_it e_it = e.get_iterator();
    delete *e_it;
    edge_list.erase(e_it);
  }
  
  void hide_edge(Edge_it e)
  {
    Edge_list_it e_it = e.get_iterator();
    (**e_it).hide();
  }
  
  void restore_edge(Edge_it e)
  {
    Edge_list_it e_it = e.get_iterator();
    (**e_it).restore();
  }
  
  int num_of_nodes()
  {
    return node_list.size();
  }
  
  int degree(Node_it n)
  {
    Node_list_it n_it = n.get_iterator();
    return (**n_it).degree();
  }
  
  Node_it nodes_begin()
  {
    return Node_it(node_list.begin());
  }
  
  Node_it nodes_end()
  {
    return Node_it(node_list.end());
  }
  
  Adj_node_it adj_nodes_begin(Node_it n)
  {
    Node_list_it n_it = n.get_iterator();
    Adj_node_da da(n_it);
    return Adj_node_it((**n_it).adj_edges_begin(), da);
  }
  
  Adj_node_it adj_nodes_end(Node_it n)
  {
    Node_list_it n_it = n.get_iterator();
    return Adj_node_it((**n_it).adj_edges_end());
  }
  
  Node_it opposite(Node_it n, Edge_it e)
  {
    Node_list_it n_it = n.get_iterator();
    Edge_list_it e_it = e.get_iterator();
    return Node_it((**e_it).opposite(n_it));
  }
  
  Node_it get_node_it(Adj_node_it n)
  {
    Adj_node_da n_da = n.get_da();
    return n_da.get_node();
  }
  
  int num_of_edges()
  {
    return edge_list.size();
  }
  
  Edge_it edges_begin()
  {
    return Edge_it(edge_list.begin());
  }
  
  Edge_it edges_end()
  {
    return Edge_it(edge_list.end());
  }
  
  Adj_edge_it adj_edges_begin(Node_it n)
  {
    Node_list_it n_it = n.get_iterator();
    return Adj_edge_it((**n_it).adj_edges_begin());
  }
  
  Adj_edge_it adj_edges_end(Node_it n)
  {
    Node_list_it n_it = n.get_iterator();
    return Adj_edge_it((**n_it).adj_edges_end());
  }
  
  Edge_it get_edge_it(Adj_edge_it e)
  {
    Adj_edge_list_it e_it = e.get_iterator();
    return Edge_it(*e_it);
  }
    
private:
  Node_list node_list;
  Edge_list edge_list;
};
#endif
